=begin

  *Name: atlas-SNPEvaluate.rb 
  *Description:
    'atlasSNPEvaluate.rb' is designed to evalute the posterior probabilities of putative SNP sites. The evaluation results (Posterior(SNP|Sj,c)j) are appended to the last column of putative SNP list output file, which is produced by "atlasSubstitution.rb". The current release can work on data from 454 FLX and XLR.
  *Authors: Yufeng Shen, Zhengzheng Wan, Cristian Coarfa and Fuli Yu @ BCM - Human Genome Sequencing Center
  *Date:  07/29/2009

=end


#!/usr/bin/env ruby
require 'getoptlong'
require 'bigdecimal'


class AtlasEvaluate
#Initialize options
def AtlasEvaluate.processArguments()
opts = GetoptLong.new(
    ["--input", "-i", GetoptLong::REQUIRED_ARGUMENT],
    ["--help", "-h", GetoptLong::NO_ARGUMENT],
    ["--XLR", "-x", GetoptLong::NO_ARGUMENT],
    ["--output","-o", GetoptLong::REQUIRED_ARGUMENT],
    ["--priorerror","-e", GetoptLong::OPTIONAL_ARGUMENT]
)

optHash = {}
opts.each do |opt, arg|
  optHash[opt] = arg
end

AtlasEvaluate.usage() if (optHash.key?("--help"));

AtlasEvaluate.usage() if (optHash.empty?);
return optHash
end

#Usage information
def AtlasEvaluate.usage(msg='')
  unless (msg.empty?)
        puts "\n#{msg}\n"
  end
  puts "
  
Program Overview:

'atlasSNPEvaluate.rb' is designed to evalute the posterior probabilities of putative SNP sites. The evaluation results (Posterior(SNP|Sj,c)j) are appended to the last column of putative SNP list output file, which is produced by 'atlasSubstitution.rb'. The current release can work on data from 454 FLX and XLR.
  

Usage:
ruby atlasSNPEvaluate.rb -i [putative SNP list] -x [option for XLR data] -e [priorerror] -o [output]
  
Required arguments:
  -i,--input         putative SNP list
  -o,--output        output file name
  
Optional arguments:
  -x,--XLR           for 454 XLR data when appended; otherwise for 454 FLX
  -e,--priorerror    to assign Prior(error|c) when the variant read coverage is greater than 2. Default=0.1
  
Example #1. To evaluate the posterior SNP probabilities for the input 'chr1.xm.SNP.list' from 454 XLR platform, when Prior(Error|Cov)=0.4 and the variant read coverage > 2. The output file is named as 'chr1.xm.SNP.list.eva.prior_0.4'. 
     
     ruby atlasSNPEvaluate.rb -i chr1.xm.SNP.list -x -e 0.4 -o chr1.xm.SNP.list.eva.prior_0.4
     
Example #2. To evaluate the posterior SNP probabilities for the input 'chr1.xm.SNP.list' from 454 FLX platform, when Prior(error|c)=0.1 and the variant read coverage > 2. The output file is named as 'chr1.xm.SNP.list.eva.prior_0.1'.
     
     ruby atlasSNPEvaluate.rb -i chr1.xm.SNP.list -o chr1.xm.SNP.list.eva.prior_0.1"
  exit(2);
end

#Initialize the object of class 'AtlasEvaluate'
def initialize(optHash)
    @optHash=optHash
    setParameters()
end

#Logistic regression coefficients
$logistic_FLX={:intercept=>-5.36, :quality_score=>0.11, :swap=>-4.88, :NQS_pass=>1.07, :relpos=>0.91}
$logistic_XLR={:intercept=>-3.31, :quality_score=>0.1, :NQS_pass=>0.26, :swap=>-3.5, :relpos=>-0.37}

#Sj (variant read coverage <= 2)initialization
$errorPrior_test_less_3=[0.867, 0.867, 0.062, 0.062, 0.063, 0.063, 0.008, 0.008, 0.0003, 0.0003]
$snpPrior_test_less_3=[0.78, 0.78, 0.122, 0.122, 0.049, 0.049, 0.049, 0.049, 0.00001, 0.00001]

#Sj (variant read coverage >= 3)initialization
$errorPrior_test_greater_3=[0.676, 0.676, 0.102, 0.102, 0.074, 0.074, 0.028, 0.028, 0.12, 0.12]
$snpPrior_test_greater_3=[0.00001, 0.00001, 0.006, 0.006, 0.006, 0.006, 0.041, 0.041, 0.947, 0.947]

#Initialize the parameters
def setParameters()

if (!@optHash.key?('--input') or !@optHash.key?('--output'))
  AtlasEvaluate.usage("Option missing!")
  exit(2);
end

if @optHash.key?("--XLR")
  @XLR=true
  @logistic=$logistic_XLR
else
  @XLR=false
  @logistic=$logistic_FLX
end

if @optHash.key?("--priorerror")
  @perror_highcov=@optHash["--priorerror"]
  @psnp_highcov=1-@perror_highcov.to_f
else
  @perror_highcov=0.1
  @psnp_highcov=0.9
end

$stderr.puts "Prior(error|c): "+@perror_highcov.to_s
$stderr.puts "Prior(SNP|c):"+@psnp_highcov.to_s

#Create output file writer
@outputFile=@optHash["--output"]
@missedFile=@optHash["--output"]+".missed_reads"
@subsFile=@optHash["--input"]
end

#Evaluate the substitutions
def evaluate()
  @errPrior=[] 
  @snpPrior=[]
  prior_error_c=-10
  prior_snp_c=-10
  var_cov=0
  outputWriter=File.new(@outputFile,"w")
  subsReader=File.open(@subsFile,"r")
  missedWriter=File.new(@missedFile,"w")
  outputWriter.print "refName\tcoordinate\trefBase\thomopolymer\trefEnv\ttotalCoverage\tvariantBase\toriQual\tvariantReadCov\tnumAlternativeReads\treadsInfo\tPr(error)j\tPr(SNP)j\tPr(Sj|error,c)\tPr(Sj|SNP,c)\tPrior(error|c)\tPrior(SNP|c)\tPosterior(SNP|Sj,c)j\n"
  subsReader.each do  |line|
  cols=line.split(/\s+/)

  coord, homo,totalCov,varbase,varCov,readinfo =cols[1].to_i, cols[3].to_i,cols[5].to_i, cols[6], cols[9].to_i, cols[11]
  
  pr_error_j=1
  
  next if varbase=="variantBase"
  var_cov=0
  rstr=""
  #lp_str=""
  readinfo.split(";").each do |r|
    if r=~ /^(\S+)\((\d+)\)(\S+)\((\d+)\)\S+([+|-])\S+\((\S+)\/(\S+)\/(\d+)\)\((\d+)\/(\d+)\)(\S+)/
      
      base=$1
        qual=$2.to_i
        name=$3
        dist=$4.to_i
        dir=$5
        sub=$6.to_f
        indel=$7.to_f
        tail=$8.to_i
        nqs=$9.to_i
        len=$10
        swap=$11
        relpos=dist/len.to_f
        
        next if base != varbase
        next if nqs==nil or len==nil
        
        if base == 'N'
          logOdd = 0
          errorPosteriorFull = 1
        else
          if swap =="swap"
            swap = 1
            elsif swap =="mnp"
            swap = 0.5
            else
            swap = 0
          end
          
          #Logistic regression p-value
          predict = @logistic[:intercept] + @logistic[:quality_score]*qual + @logistic[:swap]*swap + @logistic[:NQS_pass]*nqs + @logistic[:relpos] * relpos
          logisticP = Math.exp(predict) / (1 + Math.exp(predict))
          
          pr_error_i=1-logisticP
          pr_error_j*=pr_error_i
          
        end
        
        var_cov+=1
        
        #lp_str<<pr_error_i.to_s+"|"
        rstr<<r+"("+pr_error_i.to_s+");"
    end
  end
  
  if var_cov==0
    missedWriter.puts line
  end
  
  #Select Sj
  if var_cov>0
    outputWriter.print cols[0..6].join("\t") + "\t"+cols[8..10].join("\t")+"\t"+rstr+"\t" #Print the existing columns
    
    #Prior(SNP)j="Sj"
    pr_snp_j=1-BigDecimal(pr_error_j.to_s)
    bin = (BigDecimal(pr_snp_j.to_s) *10).floor                           #Determine the bin of Sa
  
    if bin==10
      bin=9
    end
    
    #Determine Sj sets according to variant reads coverage.
    if var_cov<3
      @errPrior=$errorPrior_test_less_3                                     #Select Sj sets for Errors
      @snpPrior=$snpPrior_test_less_3                                      #Select Sj sets for SNPs
      prior_error_c=0.9                                                              #Prior(Error|Cov)
      prior_snp_c=0.1                                                                #Prior(SNP|Cov)
    elsif var_cov>=3
      @errPrior=$errorPrior_test_greater_3
      @snpPrior=$snpPrior_test_greater_3
      
      prior_error_c=@perror_highcov.to_f
      prior_snp_c=@psnp_highcov.to_f
    end
  
  #Get Pr(Sj|error,c) and Pr(Sj|SNP,c)
  pr_Sj_error_c = @errPrior[bin]
  pr_Sj_SNP_c = @snpPrior[bin]
  
  #For debugging
  if pr_Sj_error_c==nil or pr_Sj_SNP_c==nil
    $stderr.puts bin
    $stderr.puts prior_error_c
    $stderr.puts prior_snp_c
    $stderr.puts line
  end

  
  errorPosterior = pr_Sj_error_c * prior_error_c
  snpPosterior = pr_Sj_SNP_c * prior_snp_c
  
  #Get Posterior(SNP|Sj,Cov)j
  pr_SNP_Sj_c_j = 1.0 / (1 + errorPosterior/snpPosterior)
  
  #print the useful variables and final Posterior(SNP|Sj,Cov)j
  outputWriter.print "#{pr_error_j}\t#{pr_snp_j}\t#{pr_Sj_error_c}\t#{pr_Sj_SNP_c}\t#{prior_error_c}\t#{prior_snp_c}\t#{pr_SNP_Sj_c_j}\n"
  end
  end
outputWriter.close
missedWriter.close
subsReader.close
end

end

optHash=AtlasEvaluate.processArguments()
evaluater=AtlasEvaluate.new(optHash)
evaluater.evaluate()
