=begin

  *Name: nqsPass.rb 
  *Description:
    'nqsPass.rb' is designed to calculate whether a variant base passes the Neighborhood Quality Standard (NQS) test. A base is defined as passing NQS test ( p=1) if its own phred-like quality score is above or equal to 20 and quality scores of the five flanking bases on both sides are all above or equal to 15.
  *Authors: Yufeng Shen, Zhengzheng Wan, Cristian Coarfa and Fuli Yu @ BCM - Human Genome Sequencing Center
  *Date:  07/29/2009

=end


#!/usr/bin/env ruby
require 'getoptlong'
require 'bigdecimal'


class  NqsPass

#Initialize options 
def NqsPass.processArguments()

opts = GetoptLong.new(
    ["--qual", "-q", GetoptLong::REQUIRED_ARGUMENT],
    ["--xm", "-x", GetoptLong::REQUIRED_ARGUMENT],
    ["--help","-h", GetoptLong::NO_ARGUMENT],
    ["--outputPrefix","-o", GetoptLong::REQUIRED_ARGUMENT]
)

optHash = {}
opts.each do |opt, arg|
  optHash[opt] = arg
end

NqsPass.usage() if (optHash.key?("--help"));

NqsPass.usage() if (optHash.empty?);
return optHash
end

#The usage information of the program
def NqsPass.usage(msg='')
    unless (msg.empty?)
        puts "\n#{msg}\n"
    end
    puts "
    
Program Overview:

'nqsPass.rb' is designed to calculate whether a variant base passes the Neighborhood Quality Standard (NQS) test. A base is defined as passing NQS test ( p=1) if its own phred-like quality score is above or equal to 20 and quality scores of the five flanking bases on both sides are all above or equal to 15.


Usage: 
nqsPass.rb -q [fasta quality file] -x [crossmatch file] -o [prefix of output file]

Required arguments:

  -q,--qual               fasta quality score file
  -x,--xm                 crossmatch file
  -o,--outputPrefix       user-defined output prefix, the full filename will be [prefix].NQS

Output format:
  read_name[tab]read_length[tab]distance_to_3'[tab]quality[tab]NQS_pass

Example:
  Taking fasta quality file 'fasta.part_0_1.fa.qual' and crossmatch file 'chr1.xm.part_0'
  as inputs, to compute NQS and print results to an output file named as 'part_0_1.NQS'.
  
  ruby nqsPass.rb -i fasta.part_0_1.fa.qual -x chr1.xm.part_0 -o part_0_1
  
  "
   exit(2);
end

def initialize(optHash)
    @optHash=optHash
    setParameters()
end

#Initialize the parameters
def setParameters()
    if (!@optHash.key?('--qual') or !@optHash.key?("--xm") or !@optHash.key?("--outputPrefix") )then
        NqsPass.usage("Option missing!")
        exit(2);
    end
    
    @xmFile=@optHash["--xm"]
    @fastaFile=@optHash["--qual"]
    #@outputRoot=@optHash["--outputFileRoot"]
    @outputPrefix=@optHash["--outputPrefix"]
    
    @file_name=File.basename(@fastaFile)
    @file_path=File.dirname(@fastaFile)

    
    @file_NQS=@outputPrefix+'.NQS'
end

$pattern=/^\s*(\d+)\s+(\d+\.\d+)\s+([0-9\.]+)\s+([0-9\.]+)\s+(\S+)\s+(\d+)\s+(\d+)\s+\((\d+)\)\s+(C*?)\s*?(\S+)\s+(\S+)\s+(\d+)\s+(\S+)\s+(\**?)/

#Parse out the variation position from crossmatch file
def extractVariationPosition()
    @hash_fasta_dist={}
    query,qsize,score= '',0, 0
    sub= 0
    xmReader=File.open(@xmFile,"r")
    xmReader.each do |line|
    if line.match($pattern)
        score,sub,del,ins,query,qstart,qend,qright,tstrand,target,d1,d2,d3,lab =$1.to_f,$2.to_f,$3.to_f,$4.to_f,$5,$6.to_i,$7.to_i,$8.to_i,$9,$10,$11,$12,$13,$14
        qsize = qend + qright
   
        elsif line=~ /^\s(\S+)\s+(\d+)\s+(\S+)\((\d+)\)\s+(\d+)\s+(\S+)/
          type, qplace, tplace = $1, $2.to_i, $5.to_i
    
          if type =~ /^S/                                   # substitution
            dist = qsize - qplace                          #calcuate distance to 3'
            if @hash_fasta_dist[query]==nil
                @hash_fasta_dist[query]=[]
            end
            @hash_fasta_dist[query]<<dist 
          end
        end
    end
    $stderr.puts @hash_fasta_dist.size
    xmReader.close
end

#Determine 'NQS pass' status.
def calculateNQSPass()
    @hash_nqs={}
    p_status=''
    name='1'
    qual=[]
    count=0
   fastaReader=File.open(@fastaFile,"r")
   puts @fastaFile
   
   fastaReader.each do |line|
     if line=~ /^>(\S+)/
        if @hash_fasta_dist[name]!=nil
                @hash_fasta_dist[name].uniq.sort.each do |i|
                    dist=i.to_i
                    #puts dist
                    pos=count-1-dist                       #Calculate the base position
                    if count>10                               #Read length >= 11
                        if dist <5                              #Last four bases
                            p_status="0"                    #Set NQS pass to be "0", failed
                        elsif pos<5                           #The first four bases
                            p_status="1"                   #Set NQS pass to be "1", passed
                        else 
                            if qual[pos].to_i>=20 and qual[pos-1].to_i>=15 and qual[pos-2].to_i>=15 and qual[pos-3].to_i>=15 and qual[pos-4].to_i>=15 and qual[pos-5].to_i>=15 and qual[pos+1].to_i>=15 and qual[pos+2].to_i>=15 and qual[pos+3].to_i>=15 and qual[pos+4].to_i>=15 and qual[pos+5].to_i>=15
                                p_status="1"               #Set NQS pass to be "1", passed
                            else 
                                p_status="0"              #Set NQS pass to be "0", failed
                            end
                        end
                    else                                       #Read length <= 10
                        p_status="1"                      #Set NQS pass to be "1", passed
                    end
                    
                    #Print the distance to 3' (pos), "NQS pass", Quality and Read length into a string
                    str=dist.to_s+'.'+p_status+'.'+qual[pos].to_s+'.'+count.to_s
                
                    @hash_nqs[name]<<str
                end
           
        end
        ##Initialize the hash table for storing NQS information
        name=$1                                           #Read name
        next if @hash_fasta_dist[name]==nil   #Ignore those reads not existing in .dist file
        if @hash_fasta_dist[name]!=nil 
            if @hash_nqs[name]==nil
                @hash_nqs[name]=[]
            end
        end
        count=0                                             #Counter of read length
        qual=[]                                             #array of quality scores
    else
        next if @hash_fasta_dist[name]==nil
        if @hash_fasta_dist[name]!=nil
            #start to write to quality array
            cols=line.split(/\s+/)
            count+=cols.length                        #Adding to counter
            
            qual.concat(cols)
        end
    end  
   end
   if @hash_fasta_dist[name]!=nil
             #   $stderr.puts hash_fasta_dist[name].size
                @hash_fasta_dist[name].uniq.sort.each do |i|
                    dist=i.to_i
                    pos=count-1-dist
                    if count>10
                        if dist <5
                            p_status="0"
                        elsif pos<5
                            p_status="1"
                        else
                            pos_center=qual[pos].to_i
                            if qual[pos].to_i>=20 and qual[pos-1].to_i>=15 and qual[pos-2].to_i>=15 and qual[pos-3].to_i>=15 and qual[pos-4].to_i>=15 and qual[pos-5].to_i>=15 and qual[pos+1].to_i>=15 and qual[pos+2].to_i>=15 and qual[pos+3].to_i>=15 and qual[pos+4].to_i>=15 and qual[pos+5].to_i>=15
                            
                                p_status="1"
                            else
                                p_status="0"
                            end
                        end
                    else
                        p_status="1"
                    end
                    str=dist.to_s+'.'+p_status+'.'+qual[pos].to_s+'.'+count.to_s
                    
                    @hash_nqs[name]<<str
                end
            #end
        end
   $stderr.puts @hash_nqs.size
   fastaReader.close
end

#Print result
def printResult()
    nqsWriter=File.new(@file_NQS,"w")
    nqsWriter.print "readName\treadLength\tdistanceToRightEnd\tQuality\tnqsPASS\n"
    @hash_nqs.keys.sort.each do |n|
        @hash_nqs[n].sort.each do |i|
            str=i
            if str=~ /(\S+).(\S+)\.(\S+)\.(\S+)/
              d=$1
              p=$2
              qual=$3
              length=$4
            end
            nqsWriter.print "#{n}\t#{length}\t#{d}\t#{qual}\t#{p}\n"
        end
    end
    nqsWriter.close
end
end

#Main methods
optHash=NqsPass.processArguments()
nqsPassDeterminter=NqsPass.new(optHash)
nqsPassDeterminter.extractVariationPosition()
nqsPassDeterminter.calculateNQSPass()
nqsPassDeterminter.printResult()
exit(0);